import os
import pickle
from typing import Dict, List, Tuple
import numpy as np
from sklearn.metrics import accuracy_score
from sklearn.mixture import GaussianMixture
from sklearn.model_selection import StratifiedKFold
from sklearn import svm

from data import (
    PIPELINE_PICKLE,
    TRAIN_DATA_PICKLE,
    RANDOM_STATE as DATA_RANDOM_STATE,
    image_to_vector,
    wav_to_mfcc,
    mfcc_to_vector,
    pairwise_pseudo_augment,
    create_pickles_dir,
)


# ------------------ constants ------------------
RANDOM_STATE = DATA_RANDOM_STATE
CLASSIFIERS_PICKLE = "pickles/classifiers.pkl"


# ------------------ classifiers ------------------
class GMMClassifier:
    def __init__(
            self,
            components_list = [1, 2, 4, 8],
            cov_types = ["diag", "full"],
            reg_covar = 1e-6,
            random_state = RANDOM_STATE,
            verbose = False,
        ):
        self.components_list = components_list
        self.cov_types = cov_types
        self.reg_covar = reg_covar
        self.random_state = random_state
        self.verbose = verbose
        self.classes = None

    def _fit_single(self, k, cov, X, y):
        models: Dict[str, GaussianMixture] = {}
        for label in np.unique(y):
            gm = GaussianMixture(
                n_components=k,
                covariance_type=cov,
                random_state=self.random_state,
                reg_covar=self.reg_covar,
            )
            gm.fit(X[y == label])
            models[label] = gm
        return models

    def _predict(self, models, X):
        classes = np.array(list(models))
        log_lh = np.column_stack([models[cls].score_samples(X) for cls in classes])
        return classes[np.argmax(log_lh, axis=1)]

    def fit(self, X: np.ndarray, y: np.ndarray):
        best_acc = -1.0
        best_params = None
        skf = StratifiedKFold(n_splits=3, shuffle=True, random_state=0)

        # Search best params by fitting on single class
        for k in self.components_list:
            for cov in self.cov_types:
                accs: List[float] = []
                try:
                    for tr_idx, val_idx in skf.split(X, y):
                        # Skip if not enough samples for k components
                        tr_y = y[tr_idx]
                        if (np.bincount(np.searchsorted(np.unique(y), tr_y)) < k).any():
                            raise ValueError("Not enough samples for k components")
                        models = self._fit_single(k, cov, X[tr_idx], tr_y)
                        preds = self._predict(models, X[val_idx])
                        accs.append(accuracy_score(y[val_idx], preds))
                except ValueError as e:
                    if self.verbose: print(f"Skipping params (k={k}, cov={cov}) due to fitting error: {e}")
                    continue

                mean_acc = np.mean(accs)
                if mean_acc > best_acc:
                    best_acc = mean_acc
                    best_params = (k, cov)

        # Final training
        if best_params is None:
            raise RuntimeError("No valid config found")
        k_opt, cov_opt = best_params
        if self.verbose: print(f"Best params: n_components={k_opt}, cov_type={cov_opt}")
        self.models_ = self._fit_single(k_opt, cov_opt, X, y)
        self.classes = np.array(list(self.models_))
        return self

    def predict(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        log_lh = np.column_stack([self.models_[cls].score_samples(X) for cls in self.classes])
        preds = self.classes[np.argmax(log_lh, axis=1)]
        return preds, log_lh
    

class SVMClassifier:
    def __init__(
        self,
        C_list = [0.1, 1, 5, 10],
        kernels = ["linear", "rbf"],
        cv_splits = 3,
        random_state = RANDOM_STATE,
        verbose = False,
    ):
        self.C_list = C_list
        self.kernels = kernels
        self.cv_splits = cv_splits
        self.random_state = random_state
        self.verbose = verbose
        self.classes = None

    def fit(self, X: np.ndarray, y: np.ndarray):
        best_acc = -1.0
        best_params = None
        skf = StratifiedKFold(n_splits=self.cv_splits, shuffle=True, random_state=self.random_state)

        # Search best params by fitting on single class
        for C in self.C_list:
            for kern in self.kernels:
                accs: List[float] = []
                try:
                    for tr_idx, val_idx in skf.split(X, y):
                        model = svm.SVC(C=C, kernel=kern, probability=False, random_state=self.random_state)
                        model.fit(X[tr_idx], y[tr_idx])
                        preds = model.predict(X[val_idx])
                        accs.append(accuracy_score(y[val_idx], preds))
                except Exception as e:
                    if self.verbose: print(f"Skipping params (C={C}, kernel='{kern}') due to fitting error: {e}")
                    continue

                mean_acc = np.mean(accs)
                if mean_acc > best_acc:
                    best_acc = mean_acc
                    best_params = (C, kern)

        if best_params is None:
            raise RuntimeError("No valid config found")

        C_opt, kern_opt = best_params
        if self.verbose: print(f"Best params: C={C_opt}, kernel='{kern_opt}'")

        # final training
        self.best_estimator_ = svm.SVC(C=C_opt, kernel=kern_opt, probability=False, random_state=self.random_state)
        self.best_estimator_.fit(X, y)
        self.classes = np.unique(y)
        return self

    def predict(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        preds = self.best_estimator_.predict(X)
        llh = self.best_estimator_.decision_function(X)
        return preds, llh


# ------------------ vectorization utilities ------------------
def vectorize_image(path: str, pipeline: dict) -> np.ndarray:
    raw = image_to_vector(path)
    return pipeline["pca_img"].transform(pipeline["scaler_img"].transform(raw.reshape(1, -1)))[0]


def vectorize_audio(path: str, pipeline: dict) -> np.ndarray:
    mfcc_frames = wav_to_mfcc(path)
    vec = mfcc_to_vector(mfcc_frames, pipeline["scaler_aud_frames"], pipeline["pca_aud_frames"])
    return pipeline["scaler_aud"].transform(vec.reshape(1, -1))[0]

TEST_DATA = False

# ------------------ main ------------------
if __name__ == "__main__":
    # Load data
    with open(TRAIN_DATA_PICKLE, "rb") as f:
        data = pickle.load(f)
        train = data["train"]
        if TEST_DATA:
            test_paths = data["test_paths"]

    with open(PIPELINE_PICKLE, "rb") as f:
        pipeline = pickle.load(f)

    if TEST_DATA:
        # Vectorize test data with pipeline
        X_img_test = np.vstack([vectorize_image(path, pipeline) for path in test_paths["img"]])
        y_img_test = np.array([os.path.basename(os.path.dirname(path)) for path in test_paths["img"]])

        X_aud_test = np.vstack([vectorize_audio(path, pipeline) for path in test_paths["aud"]])
        y_aud_test = np.array([os.path.basename(os.path.dirname(path)) for path in test_paths["aud"]])

        X_mm_test, y_mm_test = pairwise_pseudo_augment(X_img_test, y_img_test, X_aud_test, y_aud_test)
        X_mm_test = pipeline["pca_mm"].transform(pipeline["scaler_mm"].transform(X_mm_test))

    # Training
    print("-------- GMM --------")
    gmm_aud = GMMClassifier().fit(train["X_aud"], train["y_aud"])
    if TEST_DATA:
        y_pred_aud, _ = gmm_aud.predict(X_aud_test)
        acc_aud = accuracy_score(y_aud_test, y_pred_aud)
        print(f"audio acc:      {acc_aud:.3f}")

    gmm_img = GMMClassifier().fit(train["X_img"], train["y_img"])
    if TEST_DATA:
        y_pred_img, _ = gmm_img.predict(X_img_test)
        acc_img = accuracy_score(y_img_test, y_pred_img)
        print(f"image acc:      {acc_img:.3f}")

    gmm_mm = GMMClassifier().fit(train["X_mm"], train["y_mm"])
    if TEST_DATA:
        y_pred_mm, _ = gmm_mm.predict(X_mm_test)
        acc_mm = accuracy_score(y_mm_test, y_pred_mm)
        print(f"multimodal acc: {acc_mm:.3f}")


    print("\n-------- SVM --------")
    svm_aud = SVMClassifier().fit(train["X_aud"], train["y_aud"])
    if TEST_DATA:
        y_pred_aud_svm, _ = svm_aud.predict(X_aud_test)
        acc_aud_svm = accuracy_score(y_aud_test, y_pred_aud_svm)
        print(f"audio acc:      {acc_aud_svm:.3f}")

    svm_img = SVMClassifier().fit(train["X_img"], train["y_img"])
    if TEST_DATA:
        y_pred_img_svm, _ = svm_img.predict(X_img_test)
        acc_img_svm = accuracy_score(y_img_test, y_pred_img_svm)
        print(f"image acc:      {acc_img_svm:.3f}")

    svm_mm = SVMClassifier().fit(train["X_mm"], train["y_mm"])
    if TEST_DATA:
        y_pred_mm_svm, _ = svm_mm.predict(X_mm_test)
        acc_mm_svm = accuracy_score(y_mm_test, y_pred_mm_svm)
        print(f"multimodal acc: {acc_mm_svm:.3f}")


    # Pickle classifiers
    create_pickles_dir()
    with open(CLASSIFIERS_PICKLE, "wb") as f:
        pickle.dump({
            "gmm": {
                "aud": gmm_aud,
                "img": gmm_img,
                "mm": gmm_mm,
            },
            "svm": {
                "aud": svm_aud,
                "img": svm_img,
                "mm": svm_mm,
            },
        }, f)
    print("\nClassifiers pickled.")