import os
import sys
import pickle
import argparse
import numpy as np

from data import (
    PIPELINE_PICKLE,
)
from classifier_training import (
    CLASSIFIERS_PICKLE,
    vectorize_image,
    vectorize_audio,
    # required for unpickling
    GMMClassifier,
    SVMClassifier,
)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Classify person from image, audio, or both using pretrained GMM or SVM classifier",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    clf_type_group = parser.add_mutually_exclusive_group(required=True)
    clf_type_group.add_argument("--gmm", action="store_true", help="Use GMM classifier")
    clf_type_group.add_argument("--svm", action="store_true", help="Use SVM classifier")

    modality_group = parser.add_mutually_exclusive_group(required=True)
    modality_group.add_argument("--img", action="store_true", help="Image modality")
    modality_group.add_argument("--aud", action="store_true", help="Audio modality")
    modality_group.add_argument("--mm",  action="store_true", help="Multimodal mode")
    parser.add_argument(
        "paths",
        nargs="+",
        help="Path(s) to input file(s): 1 for img/aud, img & aud for mm",
    )

    args = parser.parse_args()

    clf_type = "gmm" if args.gmm else "svm"
    if args.img:
        modality = "img"
    elif args.aud:
        modality = "aud"
    else:
        modality = "mm"

    if modality in ["img", "aud"] and len(args.paths) != 1:
        parser.error(f"Modality '{modality}' requires 1 input file path")
    if modality == "mm" and len(args.paths) != 2:
        parser.error("Multimodal mode requires 2 paths (image & audio)")

    if not os.path.exists(PIPELINE_PICKLE):
        sys.exit(f"Pipeline file '{PIPELINE_PICKLE}' not found")
    with open(PIPELINE_PICKLE, "rb") as f:
        pipeline = pickle.load(f)

    # load classifiers
    if not os.path.exists(CLASSIFIERS_PICKLE):
        sys.exit(f"Classifier file '{CLASSIFIERS_PICKLE}' not found")
    with open(CLASSIFIERS_PICKLE, "rb") as f:
        all_clfs = pickle.load(f)

    clf = all_clfs[clf_type][modality]

    file_name = args.paths[0].split(".")[0]
    file_name = file_name.split("/")[1].strip()

    # vectorize inputs
    if modality == "img":
        vec = vectorize_image(args.paths[0], pipeline)
    elif modality == "aud":
        vec = vectorize_audio(args.paths[0], pipeline)
    else:
        vec_img = vectorize_image(args.paths[0], pipeline)
        vec_aud = vectorize_audio(args.paths[1], pipeline)
        concat = np.concatenate([vec_img, vec_aud]).reshape(1, -1)
        vec = pipeline["pca_mm"].transform(
            pipeline["scaler_mm"].transform(concat)
        )[0]

    # predict
    pred, llh = clf.predict(vec.reshape(1, -1))
    pred, llh = pred[0], llh[0]

    classes_default_order = clf.classes.astype(int)
    classes_sorted = sorted(classes_default_order)
    
    llh_sorted = [llh[np.where(classes_default_order == cls)[0][0]] for cls in classes_sorted]
    if clf_type == "svm":
        llh_sorted = ["NaN"] * 31
        print(f"{file_name} {pred}", " ".join([f"{l}" for l in llh_sorted]))
    else:
        print(f"{file_name} {pred}", " ".join([f"{l:.3f}" for l in llh_sorted]))