import glob
import os
import pickle
from typing import List, Tuple, Dict

import numpy as np
import pandas as pd
from skimage.io import imread
from scipy.io import wavfile
from sklearn.decomposition import PCA
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler

from ikrlib import mfcc

# ------------- constants -------------
DATA_ROOT = "SUR_projekt2024-2025"
PIPELINE_PICKLE = "pickles/pipeline.pkl"
TRAIN_DATA_PICKLE = "pickles/train_data.pkl"
TEST_SPLIT_RATIO = 0.0
RANDOM_STATE = 42


# ------------- helpers -------------
def create_pickles_dir():
    if not os.path.exists("pickles"):
        os.makedirs("pickles")

def list_files(pattern: str) -> List[str]:
    """Sorted list of files matching pattern"""
    return sorted(glob.glob(pattern, recursive=True))


def person_id_from_path(path: str) -> str:
    return os.path.basename(os.path.dirname(path))


# ------------- image -------------
def image_to_vector(path: str) -> np.ndarray:
    """Flatten image to vector"""
    img = imread(path).astype(np.float32) / 255.0
    return img.ravel()


# ------------- audio -------------
def wav_to_mfcc(path: str) -> np.ndarray:
    """MFCC matrix with shape (n_frames, 13)"""
    rate, signal = wavfile.read(path)
    assert rate == 16000, f"Invalid sample rate {rate} in '{path}'"
    return mfcc(signal, 400, 240, 512, 16000, 23, 13)


def fit_audio_frame_pca(mfcc_frames_list: List[np.ndarray], n_components) -> Tuple[StandardScaler, PCA]:
    """Fit scaler + PCA on all MFCC frames"""
    frames = np.vstack(mfcc_frames_list)
    scaler = StandardScaler().fit(frames)
    frames_scaled = scaler.transform(frames)
    pca = PCA(n_components=n_components, random_state=RANDOM_STATE).fit(frames_scaled)
    return scaler, pca


def mfcc_to_vector(mfcc_frames: np.ndarray, scaler: StandardScaler, pca: PCA) -> np.ndarray:
    """PCA transform of MFCC frames and summarize with mean + variance"""
    pcs = pca.transform(scaler.transform(mfcc_frames))
    return np.concatenate([pcs.mean(0), pcs.var(0)])


# ------------- multimodal pseudo‑augmentation -------------
def pairwise_pseudo_augment(
    X_img: np.ndarray, y_img: np.ndarray,
    X_aud: np.ndarray, y_aud: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """Cartesian product of image+audio vectors with matching person_id"""
    features, labels = [], []
    for person_id in np.unique(y_img):
        i_idx = np.where(y_img == person_id)[0]
        a_idx = np.where(y_aud == person_id)[0]
        for i in i_idx:
            for j in a_idx:
                features.append(np.concatenate([X_img[i], X_aud[j]]))
                labels.append(person_id)
    return np.stack(features), np.array(labels)


# ------------- main entry‑point -------------
if __name__ == "__main__":
    # Fetch file paths
    img_paths = (
        list_files(os.path.join(DATA_ROOT, "train", "*", "*.png"))
        + list_files(os.path.join(DATA_ROOT, "dev", "*", "*.png"))
    )
    aud_paths = (
        list_files(os.path.join(DATA_ROOT, "train", "*", "*.wav"))
        + list_files(os.path.join(DATA_ROOT, "dev", "*", "*.wav"))
    )

    img_df = pd.DataFrame( {"path": img_paths, "person_id": [person_id_from_path(p) for p in img_paths]} )
    aud_df = pd.DataFrame( {"path": aud_paths, "person_id": [person_id_from_path(w) for w in aud_paths]} )


    # Stratified splitting
    if TEST_SPLIT_RATIO > 0:
        img_df_train, img_df_test = train_test_split(
            img_df,
            test_size=TEST_SPLIT_RATIO,
            stratify=img_df["person_id"],
            random_state=RANDOM_STATE,
        )
        aud_df_train, aud_df_test = train_test_split(
            aud_df,
            test_size=TEST_SPLIT_RATIO,
            stratify=aud_df["person_id"],
            random_state=RANDOM_STATE,
        )
    else:
        img_df_train, img_df_test = img_df, pd.DataFrame(columns=img_df.columns)
        aud_df_train, aud_df_test = aud_df, pd.DataFrame(columns=aud_df.columns)


    # ----- Image pipeline -----
    X_img_train = np.vstack(img_df_train["path"].apply(image_to_vector))
    scaler_img = StandardScaler().fit(X_img_train)
    X_img_train = scaler_img.transform(X_img_train)
    pca_img = PCA(n_components=0.95, random_state=RANDOM_STATE).fit(X_img_train)
    X_img_train = pca_img.transform(X_img_train)
    y_img_train = img_df_train["person_id"].values

    print(f"Image shape:      {X_img_train.shape}")

    # ----- Audio pipeline -----
    mfcc_train_list = aud_df_train["path"].apply(wav_to_mfcc).tolist()
    scaler_aud_frames, pca_aud_frames = fit_audio_frame_pca(mfcc_train_list, 0.95)
    X_aud_train = np.vstack(
        [mfcc_to_vector(m, scaler_aud_frames, pca_aud_frames) for m in mfcc_train_list]
    )
    scaler_aud = StandardScaler().fit(X_aud_train)
    X_aud_train = scaler_aud.transform(X_aud_train)
    y_aud_train = aud_df_train["person_id"].values

    print(f"Audio shape:      {X_aud_train.shape}")

    # ----- Multimodal pseudo‑augmentation -----
    X_mm_train, y_mm_train = pairwise_pseudo_augment(X_img_train, y_img_train, X_aud_train, y_aud_train)
    scaler_mm = StandardScaler().fit(X_mm_train)
    X_mm_train = scaler_mm.transform(X_mm_train)
    pca_mm = PCA(n_components=0.95, random_state=RANDOM_STATE).fit(X_mm_train)
    X_mm_train = pca_mm.transform(X_mm_train)

    print(f"Multimodal shape: {X_mm_train.shape}")


    # ----- pickling -----
    train_data = dict(
        X_img=X_img_train,
        y_img=y_img_train,
        X_aud=X_aud_train,
        y_aud=y_aud_train,
        X_mm=X_mm_train,
        y_mm=y_mm_train,
    )
    pipeline: Dict[str, object] = dict(
        scaler_img=scaler_img,
        pca_img=pca_img,
        scaler_aud_frames=scaler_aud_frames,
        pca_aud_frames=pca_aud_frames,
        scaler_aud=scaler_aud,
        scaler_mm=scaler_mm,
        pca_mm=pca_mm,
    )

    create_pickles_dir()
    with open(PIPELINE_PICKLE, "wb") as f:
        pickle.dump(pipeline, f)
    with open(TRAIN_DATA_PICKLE, "wb") as f:
        pickle.dump(
            {
                "train": train_data,
                "test_paths": {
                    "img": img_df_test["path"].tolist(),
                    "aud": aud_df_test["path"].tolist(),
                },
            },
            f,
        )
    print("\nData and pipeline pickled.")
